from graph_tool.all import *
import heapq
import networkx as nx
import read_graph as rg
import sys
import math
import random
import scipy as sci
import numpy as np
import time
import IC
from scipy.sparse.linalg import eigsh

def get_dominator_tree(G,I):
#G must be directed

    domtree=nx.DiGraph()

    num_nodes=len(G)
    dicts={}
    #using graph_tool
    g=Graph()
    vlist=g.new_vertex_property("int")
    elist_weight=g.new_edge_property("double")
    for n,nbrs in G.adjacency_iter():
        for nbr, weight in nbrs.items():
            if str(n) in dicts:
                v1=int(dicts[str(n)])
            else:
                v1=g.add_vertex()
                dicts[str(n)]=str(v1)
                vlist[v1]=n

            if str(nbr) in dicts:
                v2=int(dicts[str(nbr)])
            else:
                v2=g.add_vertex()
                dicts[str(nbr)]=str(v2)
                vlist[v2]=nbr
            log_w=-math.log(weight['weight'])
            edge=g.add_edge(v1,v2)
            elist_weight[edge]=log_w
    
    #infected node
    I0=int(dicts[str(I)])

    #run shortest path to get weights for dominator tree
    dist=shortest_distance(g,I0,target=None,weights=elist_weight)
    dist_list=dist.a
    #for value in dist_list:
        #print math.exp(-value)

    
    #get dominator tree
    dom=dominator_tree(g,I0)                
    dom_list=dom.a

    for i in range(1,len(dom_list)):
        wei=0
        if g.vertex(dom_list[i])==I0:
            wei=math.exp(-dist_list[g.vertex(i)])
        else:
            wei=math.exp(dist_list[g.vertex(dom_list[i])]-dist_list[g.vertex(i)])
        domtree.add_edge(vlist[g.vertex(dom_list[i])],vlist[g.vertex(i)],weight=(wei))
        #domtree.add_edge(vlist[g.vertex(i)],vlist[g.vertex(dom_list[i])])
     
    #for n, nbrs in domtree.adjacency_iter():
     #   for nbr in nbrs:
      #      print("%d\t%d" % (n,nbr))
    #printline(domtree)
    return domtree

def compute_nei_footprint(domtree,I):
#get the footprint of I's neighbors
    E_nei={}
    '''
    bfs=nx.bfs_successors(domtree,I)
    for key in bfs.keys():
        for nei in domtree[int(key)]:
            #print nei
            pij=domtree[int(key)][int(nei)]['weight']
            E_nei[str(nei)]=(pij*cal_footprint(domtree,nei))
    '''
    for nei in domtree[I]:
        pij=domtree[I][int(nei)]['weight']
        E_nei[str(nei)]=(pij*cal_footprint(domtree,nei))
        for node in domtree[nei]:
            pij=domtree[nei][int(node)]['weight']
            E_nei[str(node)]=(pij*cal_footprint(domtree,int(node)))
    return E_nei


def cal_footprint(G,v):
    footprint=1
    if len(G[v])==0:
        footprint=1
    else:
        for nei in G[v]:
            footprint=footprint+cal_footprint(G,nei)*G[v][nei]['weight']

    return footprint

def create_s_graph(G,node_probs):
   #add new edge, give weights 
    I=0

    G.add_node(I)
    G1=G.copy()
    #printline(G)
    for i in range(len(node_probs)):
        w=node_probs[i]
        if math.fabs(w)>10e-6:
            G1.add_edge(I,(i+1),weight=w)
    return [G1,I]

def UVDP_DAVA(filename,nodes_probs_file,K):
    #read graph
    print "UVDP Dom algorithm begin"
    g=rg.create_undirected_graph_IC(filename)
    #printline(g)
    
    #read nodes probability
    node_probs=read_nodes_probs(nodes_probs_file)
    
    #initialize the value
    #num_nodes=len(g)
    #E_nei={}
    #for i in range(num_nodes):
    #    E_nei[str(i+1)]=0
        
    start_time=time.time()    
    
    #sampling graph G_i
    [G,I]=create_s_graph(g, node_probs)
        
    #get dominator tree
    domtree=get_dominator_tree(G,I)
    #printline(domtree)
    
   #get the benefit of I's neighbors
    E_nei=compute_nei_footprint(domtree,I)

    for key in E_nei.keys():
        E_nei[key]=E_nei[key]*(1-node_probs[(int(key)-1)])
    
    #select k nodes in dava-fast
    selected_nodes=[]
    selected_nodes=heapq.nlargest(K,E_nei,key=E_nei.get)
    
    end_time=time.time()
    cost_time=end_time-start_time
    return  [selected_nodes, cost_time] 

def UVDP_netshield(filename,nodes_probs_file,K):

    #read graph
    print "UVDP net begin"
    g=rg.create_undirected_graph_IC(filename)
    #printline(g)
    
    #read infected nodes set
    node_probs=read_nodes_probs(nodes_probs_file)
    
    start_time=time.time()
    [G,I]=create_s_graph(g, node_probs)
    g=G
    #get the largest eigenvector and eigenvalue
    
    #print 'cal eigenvalue'
        #graph is the sparese matrix in scipy
    graph=nx.to_scipy_sparse_matrix(g,nodelist=range(len(g)))
    lamda,evecs_large=eigsh(graph,1,which='LM')
    u=[ evecs_large[i][0] for i in range(len(evecs_large)) ]
    #print u
    #print lamda
    #print evecs_large
    if lamda<0:
        lamda=lamda*(-1)
        
    for i in range(len(u)):
        if u[i]<0:
            u[i]=u[i]*(-1)
    #print graph.todense()
    
    for i in range(1,len(u)):
        u[i]=u[i]*(1-node_probs[(i-1)])
    

    print 'select nodes'
    #select k nodes in netshield
    S=[]
    N=len(evecs_large)
    flag=[0 for i in range(N+1)]
    prev=0

    flag[0]=1

    for i in range(K):
        #print i
        score=[-99999 for i in range(N)]
        for j in range(1,(N)):
            if flag[j]==0:
                neighbors=g.neighbors(j)
                score[j]=compute_score(S,neighbors,u,u[j],lamda,prev,flag)
       
        a=max(score)
        ii=np.argmax(score)
        S.append((ii))
        flag[ii]=1
        prev=a
    #print N
    #for i in range(K):
        #score

    end_time=time.time()
    cost_time=end_time-start_time
    return  [S, cost_time] 

def compute_score(S,neighbors,un,uj,lamda,prev,flag):
    s=prev
    for nei in neighbors:
        #if int(nei) in S:
        if flag[int(nei)]==1:
            s=s-uj*un[int(nei)]
    s=s+(2*lamda*(uj**2))
    return s

def printline(g):
    print "edgelist"
    for n, nbrs in g.adjacency_iter():
        for nbr,weight in nbrs.items():
            print("%d,%d,%f" %(n,nbr,weight['weight']))

            
# read node probs from the graph    
def read_nodes_probs(nodes_probs_file):
    node_probs=[]
    f=open(nodes_probs_file, 'r')
    lines=f.readlines()
    for line in lines:
        [id, prob]=line.split()
        node_probs.append(eval(prob))
    f.close()
    return node_probs    

def UVDP_MAX(filename,nodes_probs_file,K):

    start_time=time.time()
    
    g=rg.create_undirected_graph_IC(filename)
    node_probs=read_nodes_probs(nodes_probs_file)
    [selected_nodes,cost_time]=UVDP_DAVA(filename,nodes_probs_file,K)
    footprint=IC.IC_model(g, node_probs,200, selected_nodes)
    
    [selected_nodes1,cost_time1]=UVDP_netshield(filename,nodes_probs_file,K)
    footprint1=IC.IC_model(g, node_probs,200, selected_nodes)
    
    S=selected_nodes1
    if footprint<footprint1:
        S=selected_nodes
    
    end_time=time.time()
    cost_time=end_time-start_time
    
    return  [S, cost_time] 

def main():
    filename=sys.argv[1]
    nodes_probs_file=sys.argv[2]
    K=int(sys.argv[3])

    #filename='edgelist'
    #infected_node_file='infected_nodes'
    #K=2

    [selected_nodes,cost_time]=UVDP_MAX(filename,nodes_probs_file,K)
    print cost_time
    save_file=filename+nodes_probs_file+'_UVDP_MAX_selected_nodes'
    rg.save_selected_node(save_file,selected_nodes)
    save_file=filename+nodes_probs_file+'_UVDP_MAX_cost_time'
    rg.save_cost_time(save_file,'UVDP_MAX',cost_time)

if __name__=='__main__':
    main()
